import typing

from forest.common import proc_utils
from forest.common.parser import git_regrex_pattern
import shutil


class GitTools:

    def __init__(self, srcdir) -> None:
        self.srcdir = srcdir

    def clone(self, 
            server: str, 
            repository: str, 
            tag: str,
            proto='ssh', 
            recursive=False,
            depth=None,
            single_branch=False):

        if proto == 'ssh':
            addr = f'git@{server}:{repository}'
        elif proto == 'https':
            addr = f'https://{server}/{repository}'
        else:
            # TODO more specific exception
            raise ValueError(f'unsupported protocol "{proto}"')
        
        # create command
        # --progress flag forces progress status even if the standard 
        # error stream is not directed to a terminal.

        cmd = ['git', 'clone', '--progress']
        
        # we're asked to clone a single branch
        # note: it cannot be a commit sha1
        if single_branch:
            cmd.extend(['--branch', tag])
            cmd.append('--single-branch')

        if recursive:
            cmd.append('--recursive')

        # note: depth should imply single branch
        if depth is not None:
            cmd.extend(['--branch', tag])
            cmd.append('--single-branch')
            cmd.extend(['--depth', depth])

        cmd.extend([addr, self.srcdir])

        # clone, and delete the source folder on failure
        # (either exception or git returns != 0) 
        try:
            # Progress status is reported on the standard error stream
            clone_ok = proc_utils.call_process(args=cmd, 
                                               update_regrex_pattern=git_regrex_pattern)
            
            # checkout to requested branch/tag/commit
            if tag is not None:
                clone_ok = clone_ok and self.checkout(tag=tag)
            
            if not clone_ok:
                self.rm()
        except BaseException as e:
            # remove src and re-raise exception
            self.rm()
            raise e

        return clone_ok

    def checkout(self, tag):
        return proc_utils.call_process(['git', 'checkout', tag], cwd=self.srcdir)

    def rm(self):
        shutil.rmtree(self.srcdir,  ignore_errors=True)



from tempfile import TemporaryDirectory
import os 
from functools import partial

from forest.common import proc_utils
from forest.common.parser import make_regrex_pattern


cmake_command = 'cmake'
default_args = list()

def _construct(self, srcdir, builddir):
    self.srcdir = srcdir
    self.builddir = builddir

def _is_configured(self):

    return os.path.exists(os.path.join(self.builddir, 'Makefile'))

def _set_default_args(args):
    global default_args
    default_args = args

def _configure(self, args):

    if args is None:
        args = list()

    return _call_cmake([self.srcdir] + args + default_args, cwd=self.builddir)
    

def _build(self, target, jobs):

    return _call_cmake(['--build', self.builddir, '--target', target, '--', '-j', jobs], print_progress=True)


def _call_cmake(args, cwd='.', print_on_error=True, print_progress=False):

    args_str = list(map(str, args))
    update_regrex_pattern = make_regrex_pattern if print_progress else None
    return proc_utils.call_process(args=[cmake_command] + args_str,
                                   cwd=cwd, 
                                   print_on_error=print_on_error,
                                   update_regrex_pattern=update_regrex_pattern)


def _find_package(pkg_name: str):
        
    with TemporaryDirectory(prefix="foresttmp-") as tmpdir:

        # dir of current file
        cmake_tools_dir = os.path.dirname(os.path.abspath(__file__))
        
        # open template
        with open(os.path.join(cmake_tools_dir, 'template', 'find_package.txt'), 'r') as f:
            find_package_tpl = f.read()
        
        # configure template
        srcdir = find_package_tpl.replace('£PKG_NAME', pkg_name)

        # write it to tmp dir
        cmakelists_path = os.path.join(tmpdir, 'CMakeLists.txt')
        with open(cmakelists_path, 'w') as f:
            f.write(srcdir)
        
        return _call_cmake('.', cwd=tmpdir, print_on_error=False)




import re
import progressbar
from time import sleep
import typing


make_regrex_pattern = '\[((?:\s|\d)(?:\s|\d)\d)%\]'
git_regrex_pattern = 'Receiving objects: ((?:\s|\d)(?:\s|\d)\d)% \('


def find_progress(line: str, regrex_pattern) -> float:
    match = re.search(regrex_pattern, line)
    if match is not None:
        progress = float(match.groups()[0])
        return progress
    

def update_progress_bar(line, pbar, regrex_pattern):   
    progress = find_progress(line, regrex_pattern)
    if progress is not None:
        pbar.update(progress)


if __name__ == '__main__':
    make_example = """
Scanning dependencies of target client
Scanning dependencies of target talker
Scanning dependencies of target listener
Scanning dependencies of target malloc_example
Scanning dependencies of target server
Scanning dependencies of target homing_example
Scanning dependencies of target ros_from_rt
[  5%] Building CXX object src/clock/CMakeFiles/clock_example.dir/clock.cpp.o
[ 20%] Building CXX object src/talker_listener/CMakeFiles/talker.dir/talker.cpp.o
[ 20%] Building CXX object src/malloc_example/CMakeFiles/malloc_example.dir/malloc_example.cpp.o
[ 20%] Building CXX object src/talker_listener/CMakeFiles/listener.dir/listener.cpp.o
[ 25%] Building CXX object src/client_server/CMakeFiles/client.dir/client.cpp.o
[ 30%] Building CXX object src/client_server/CMakeFiles/server.dir/server.cpp.o
[ 35%] Building CXX object src/homing_example/CMakeFiles/homing_example.dir/homing_example.cpp.o
[ 40%] Building CXX object src/ros_from_rt/CMakeFiles/ros_from_rt.dir/ros_from_rt.cpp.o
[ 45%] Linking CXX shared library libxbotctrl_homing_example.so
[ 45%] Built target homing_example
Scanning dependencies of target joint_impedance
[ 50%] Building CXX object src/joint_impedance/CMakeFiles/joint_impedance.dir/joint_impedance.cpp.o
[ 55%] Linking CXX shared library libxbotctrl_malloc_example.so
[ 55%] Built target malloc_example
Scanning dependencies of target cartesio_rt
[ 60%] Building CXX object src/cartesio/CMakeFiles/cartesio_rt.dir/cartesio_rt.cpp.o
[ 65%] Linking CXX shared library libxbotctrl_clock_example.so
[ 65%] Built target clock_example
[ 70%] Linking CXX shared library libxbotctrl_talker.so
[ 70%] Built target talker
[ 75%] Linking CXX shared library libxbotctrl_listener.so
[ 75%] Built target listener
[ 80%] Linking CXX shared library libxbotctrl_server.so
[ 80%] Built target server
[ 85%] Linking CXX shared library libxbotctrl_client.so
[ 85%] Built target client
[ 90%] Linking CXX shared library libxbotctrl_joint_impedance.so
[ 90%] Built target joint_impedance
[ 95%] Linking CXX shared library libxbotctrl_ros_from_rt.so
[ 95%] Built target ros_from_rt
[100%] Linking CXX shared library libxbotctrl_cartesio_rt.so
[100%] Built target cartesio_rt
"""
    pbar = progressbar.ProgressBar(maxval=100, \
    widgets=[progressbar.Bar('=', '[', ']'), ' ', progressbar.Percentage()])
    pbar.start()
    for line in make_example.splitlines():
        sleep(0.1)
        update_progress_bar(line, pbar=pbar, regrex_pattern=make_regrex_pattern)
    pbar.finish()


    git_example = """
Cloning into '/home/mruzzon/alberobotics/sps_forest/src/ReflexxesTypeII'...
remote: Enumerating objects: 279, done.
remote: Counting objects: 100% (6/6), done.
remote: Compressing objects: 100% (4/4), done.
remote: Total 279 (delta 2), reused 6 (delta 2), pack-reused 273
Receiving objects:  50% (279/279), 3.35 MiB | 4.62 MiB/s, done.
Resolving deltas: 100% (109/109), done.
returned 0
calling "git checkout master"
Already on 'master'
Your branch is up to date with 'origin/master'.
returned 0
"""

    pbar = progressbar.ProgressBar(maxval=100, \
    widgets=[progressbar.Bar('=', '[', ']'), ' ', progressbar.Percentage()])
    pbar.start()
    for line in git_example.splitlines():
        sleep(0.1)
        update_progress_bar(line, pbar=pbar, regrex_pattern=git_regrex_pattern)
    pbar.finish()


import functools
import textwrap

# this is opened in main
log_file = None

class ProgressReporter:

    # recursive function call counter
    call_count = 0

    @classmethod
    def count_calls(cls, fn):
        """
        Decorator that counts the number of recursive function calls.
        This is meant to be applied to the install function
        """
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            cls.call_count += 1
            ret = fn(*args, **kwargs)
            cls.call_count -= 1
            return ret
    
        return wrapper

    @classmethod
    def print(cls, pkg, text, **kwargs):
        indent = '..' * (cls.call_count - 1)
        fmt_text = f'[{pkg}] {text}'
        fmt_text = textwrap.indent(text=fmt_text, prefix=indent)
        print(fmt_text, **kwargs)

    
    @classmethod
    def get_print_fn(cls, pkg):
        return functools.partial(cls.print, pkg)


from typing import List
import yaml
import os

from .fetch_handler import FetchHandler
from .build_handler import BuildHandler
from forest.common.eval_handler import EvalHandler
from forest.common.recipe import Cookbook


class BasicPackage:

    """
    Represent a package in its most basic form, i.e. a name
    and a list of dependencies
    """

    def __init__(self, name, depends: List[str]=None) -> None:
        self.name = name 
        self.depends = depends if depends is not None else list()
        

class Package(BasicPackage):

    """
    Represent a 'full' package, i.e. that can be cloned and built
    with git and cmake (for now that's all we support)
    """
    
    # the path to the recipe directory
    _path = None

    def __init__(self, name, depends: List[str]) -> None:
        super().__init__(name, depends=depends)
        self.fetcher = FetchHandler(name)
        self.builder = BuildHandler(name)

    @staticmethod
    def from_yaml(name, recipe) -> 'Package':
        """
        Construct a Package or BasicPackage from yaml dict

        Args:
            name (str): the package name
            recipe (dict): a dictionary with package information

        Returns:
            Package: the constructed object
        """

        # eval handler
        eh = EvalHandler.instance()

        # dependency list if any
        depends = recipe.get('depends', list())
        if depends is None:
            # empty entry "depends:" in yaml is parsed as None
            depends = list()
        depends_if = recipe.get('depends_if', dict())
        depends.extend(eh.parse_conditional_dict(depends_if))

        # create pkg
        pkg = Package(name=name, depends=depends)

        # custom fetcher and builder if we have clone/build information
        if 'clone' in recipe.keys():
            pkg.fetcher = FetchHandler.from_yaml(pkgname=name, data=recipe['clone'], recipe=recipe)

        if 'build' in recipe.keys():
            pkg.builder = BuildHandler.from_yaml(pkgname=name, data=recipe['build'], recipe=recipe)

        return pkg

    @staticmethod
    def from_file(file) -> 'Package':
        """
        Construct a Package or BasicPackage from file
        """

        name = os.path.splitext(os.path.basename(file))[0]
        with open(file, 'r') as f:
            yaml_dict = yaml.safe_load(f.read())
            if yaml_dict is None:
                yaml_dict = dict()
            return Package.from_yaml(name=name, recipe=yaml_dict)

    @staticmethod
    def from_name(name) -> 'Package':
        """
        Construct a Package or BasicPackage from its name. The recipe file
        is first fetched from the default recipe path, and the returned
        object is constructed based on its content.

        Raises:
            RuntimeError: recipe file does not exist
        """

        filename = os.path.join(Cookbook.get_recipe_path(), name + '.yaml')
        if not os.path.exists(filename):
            # TODO more specific exception
            raise FileNotFoundError(f'{filename} does not exist')
        return Package.from_file(file=filename)





import os
import re

from forest.git_tools import *
from parse import parse
# - clone repository (with given branch) into recipes/src
# - symlink all .yaml files from recipes/src/repo/tag/subdir to recipes
# - handle conflicts, as follows
#   * in default mode, choose which source to keep
#   * with --allow-recipe-overwrite, last file is used and an informative message is printed

__all__ = ['RecipeSource', 'Cookbook']


class DuplicateRecipes(Exception):
    pass


class UserInterrupt(Exception):
    pass


class RecipeSource:
    def __init__(self, server: str, username: str, repository: str, tag: str, protocol: str):
        self.server = server
        self.username = username
        self.repository = repository
        self.tag = tag
        self.protocol = protocol

    def __str__(self):
        return f"{self.server}%{self.repository}%{self.tag}"

    @classmethod
    def FromUrl(cls, url: str, tag: str, type='git', force_proto=None):

        """
        Parse git server and repository name from url.
        This is expected to be the full address to the
        git repo according to either the ssh or https
        format.

        Raises:
            ValueError: url does not match pattern
        """

        git_addr = url
        good_patterns = ['git@{}:{}/{}.git', '{}://{}/{}/{}.git']
        parse_result_ssh = parse(good_patterns[0], git_addr)
        parse_result_https = parse(good_patterns[1], git_addr)

        if parse_result_ssh is not None:
            protocol = 'ssh'
            server, username, repository = parse_result_ssh
        elif parse_result_https is not None:
            protocol, server, username, repository = parse_result_https
        else:
            raise ValueError(f'could not parse git repository from given args {url}')

        # default proto from env
        default_proto = os.getenv('HHCM_FOREST_CLONE_DEFAULT_PROTO')
        if default_proto is not None:
            protocol = default_proto

        if force_proto is not None:
            protocol = force_proto

        return cls(server=server,
                   username=username,
                   repository=repository,
                   tag=tag,
                   protocol=protocol)


# class Recipe:
#     def __init__(self, name: str, source: RecipeSource, subdir='recipes'):
#         self.name = name
#         self.subdir = subdir
#         self.source = source
#
#     def __str__(self):
#         return f"{self.source}%{self.subdir}%{self.name}"


class Cookbook:
    basedir: str
    _recipes_dirname: str = 'recipes'

    @classmethod
    def set_recipe_basedir(cls, path):
        cls.basedir = path

    @classmethod
    def get_recipe_basedir(cls):
        """
        Returns the directory with recipes inside.
        """

        if cls.basedir is None:
            raise ValueError("Recipes' folder path missing")

        return cls.basedir

    @classmethod
    def get_recipe_path(cls):
        return _find_path_to_dir(cls.get_recipe_basedir(), cls._recipes_dirname)

    @classmethod
    def get_available_recipes(cls) -> typing.List[str]:
        """
        Returns a list of available recipe names
        """

        path = cls.get_recipe_path()

        if path is None:
            return []

        files = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
        files = [os.path.splitext(f)[0] for f in files]
        files.sort()
        return files

    @classmethod
    def add_recipes(cls,
                    recipe_src: RecipeSource) -> bool:

        if os.path.exists(os.path.join(cls.get_recipe_basedir(), recipe_src.repository)):
            print(f'{recipe_src.repository}/{recipe_src.tag}: recipes source code already exists, skipping clone')
            print(f'if you want to update recipes sources (fetch/pull), you may do so using git\nin {cls.get_recipe_basedir()}')
            return False

        if os.listdir(cls.get_recipe_basedir()):
            print(f"recipes' repository already in '{cls.get_recipe_basedir()}':\nremove it if you want to add a different one")
            return False

        try:
            _clone_recipes_src(recipe_src, os.path.join(cls.get_recipe_basedir(), recipe_src.repository))

        except RuntimeError as e:
            print(e)
            return False

        return True


def _clone_recipes_src(recipe_src: RecipeSource, destination: str):
    # clone
    g = GitTools(destination)
    repo = os.path.join(recipe_src.username, recipe_src.repository) + '.git'
    if not g.clone(recipe_src.server, repo, tag=recipe_src.tag, proto=recipe_src.protocol):
        raise RuntimeError(f'{recipe_src}: recipes source code clone failed')

    print(f'{recipe_src.repository} ({recipe_src.tag}): recipes source code successfully cloned')


def _find_path_to_dir(basedir, dirname):
    for dirpath, dirs, files in os.walk(os.path.expanduser(basedir), followlinks=True):
        if dirname in dirs:
            return os.path.join(dirpath, dirname)



import subprocess
import sys
import typing
import progressbar
import os

from forest.common import print_utils
from forest.common.parser import update_progress_bar
from forest.common.forest_dirs import rootdir

call_process_verbose = False


def call_process(args: typing.List[str] = None, 
                 cwd='.', 
                 input=None,
                 verbose=False,
                 print_on_error=True, 
                 shell: bool=False,
                 timeout=None,
                 update_regrex_pattern=None
                 ) -> bool:    
    
    # convert args to string
    args = list(map(str, args))    
    
    if verbose or call_process_verbose:
        if shell:
            print(f'calling shell with command "{args}"')
        else:
            print('calling "{}"'.format(' '.join(args)))
    
    if shell:
        executable = '/bin/bash'
        popen_args = ' ' + ' '.join(args) 
    
    else:
        executable = None
        popen_args = args

    if call_process_verbose or verbose:
        # run will print output to terminal
        proc = subprocess.run(args=args, 
                        cwd=cwd, 
                        input=input, 
                        shell=shell, 
                        executable=executable)

        print(f'returned {proc.returncode}')

        return proc.returncode == 0         

    try:
        #  # universal_newlines=True equivalent to text=True (backward compatibility)
        pr = subprocess.Popen(args=popen_args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT, 
                                cwd=cwd, 
                                shell=shell, 
                                executable=executable,
                                universal_newlines=True)
        
        lines = []
        if update_regrex_pattern is not None:
            lines = _progress_bar(pr, update_regrex_pattern)
            
        if pr.wait(timeout=timeout) != 0:
            if lines:
                print_utils.log_file.writelines(lines)
            else:
                print_utils.log_file.write(pr.stdout.read())
            
            if print_on_error and not verbose:
                print_utils.log_file.seek(0)  # go to the beginning of the file
                print(print_utils.log_file.read(), file=sys.stderr)

            return False
        
        return True

    #todo: fix subprocess.CalledProcessError not working with subprocess.Popen
    except subprocess.CalledProcessError as e:
        # on error, print output (includes stderr)
        print_utils.log_file.write(e.output.decode())

        if print_on_error and not verbose:
            print(e.output.decode(), file=sys.stderr)
            
        return False

def _progress_bar(process, regrex_pattern):
    pbar = progressbar.ProgressBar(maxval=100, \
    widgets=[progressbar.Bar('=', '[', ']'), ' ', progressbar.Percentage()])
    pbar.start()
    lines = []

    while True:
        line = process.stdout.readline()
        lines.append(line)

        if not line:
            pbar.finish()
            return lines
            
        update_progress_bar(line, pbar=pbar, regrex_pattern=regrex_pattern)


def get_output(args, cwd='.', input=None, verbose=False, print_on_error=True, shell=False):

    if verbose or call_process_verbose:
        if shell:
            print(f'calling shell with command "{args}"')
        else:
            print('calling "{}"'.format(' '.join(args)))

    try:
        # check_output will not print
        out = subprocess.check_output(args=args, cwd=cwd, input=input, shell=shell)
        ret = out.decode().strip()
        if verbose or call_process_verbose:
            print('calling "{}" returned "{}"'.format(' '.join(args), ret))
        return ret
    except subprocess.CalledProcessError as e:
        # on error, print output and errors
        if print_on_error:
            print('stdout: ' + e.output.decode(), file=sys.stderr)
            print('stderr: ' + e.stderr.decode(), file=sys.stderr)
        return None



#!/usr/bin/env python3

import argparse
import getpass
import os
import sys
import argcomplete
from datetime import datetime
from forest import cmake_tools
from forest.common.eval_handler import EvalHandler

from forest.common.install import install_package, write_setup_file, write_ws_file, check_ws_file, uninstall_package, \
    clean
from forest.common.recipe import RecipeSource, Cookbook
from forest.common import sudo_refresh
from pprint import pprint

from configparser import ConfigParser
from pathlib import Path
from pkg_resources import get_distribution

# define directories for source, build, install, and recipes
from forest.common.forest_dirs import *

# just a try-except wrapper to catch ctrl+c
def main():
    try:
        if not do_main():
            print('(failed)')
            sys.exit(1)
        sys.exit(0)
    except KeyboardInterrupt:
        print('\nfailed (interrupted by user)')
        sys.exit(1)


# actual main
def do_main():
 
    # set recipe dir
    Cookbook.set_recipe_basedir(recipesdir)

    # available recipes
    available_recipes = Cookbook.get_available_recipes()
    if len(available_recipes) == 0:
        available_recipes = None

    # parse cmd line args
    buildtypes = ['None', 'RelWithDebInfo', 'Release', 'Debug']
    cloneprotos = ['ssh', 'https']
    dfl_log_file = datetime.now().strftime("/tmp/forest_%Y_%m_%d_%H_%M_%S.log")

    parser = argparse.ArgumentParser(description='forest automatizes cloning and building of software packages')
    parser.add_argument('--list', '-l', required=False, action='store_true', help='list available recipes')
    parser.add_argument('--log-file', default=dfl_log_file, help='log file for non-verbose mode')
    parser.add_argument('--verbose', '-v', required=False, action='store_true', help='print additional information')
    parser.add_argument('--version', required=False, action='store_true', help='print forest version')

    subparsers = parser.add_subparsers(dest='command')

    init_cmd = 'init'
    init_parser = subparsers.add_parser(init_cmd, help='initialize the current folder as a forest workspace')


    grow_cmd = 'grow'
    grow_parser = subparsers.add_parser(grow_cmd, help='clone, configure, and build a recipe')
    grow_parser.add_argument('recipe', nargs='?', metavar='RECIPE', choices=available_recipes, help='name of recipe with fetch and build information')
    grow_parser.add_argument('--jobs', '-j', default=1, help='parallel jobs for building')
    grow_parser.add_argument('--mode', '-m', nargs='+', required=False, help='specify modes that are used to set conditional compilation flags (e.g., cmake args)')
    grow_parser.add_argument('--config', '-c', nargs='+', required=False, help='specify configuration variables that can be used inside recipes')
    grow_parser.add_argument('--default-build-type', '-t', default=buildtypes[1], choices=buildtypes, help='build type for cmake, it is overridden by recipe')
    grow_parser.add_argument('--force-reconfigure', required=False, action='store_true', help='force calling cmake before building with args from the recipe')
    grow_parser.add_argument('--list-eval-locals', required=False, action='store_true', help='print available attributes when using conditional build args')
    grow_parser.add_argument('--clone-protocol', required=False, choices=cloneprotos, help='override clone protocol')
    grow_parser.add_argument('--clone-depth', required=False, type=int, help='set maximum history depth to save bandwidth')
    grow_parser.add_argument('--cmake-args', nargs='+', required=False, help='specify additional cmake args to be appended to each recipe (leading -D must be omitted)')
    grow_parser.add_argument('--no-deps', '-n', required=False, action='store_true', help='skip dependency fetch and build step')
    grow_parser.add_argument('--uninstall', required=False, action='store_true', help='uninstall recipe')
    grow_parser.add_argument('--clean', required=False, action='store_true', help='uninstall recipe and remove build')
    grow_parser.add_argument('--pwd', '-p', required=False, help='user password to be used when sudo permission is required (if empty, user is prompted for password); note: to be used with care, as exposing your password might be harmful!')
    grow_parser.add_argument('--verbose', '-v', required=False, action='store_true', help='print additional information')

    recipes_cmd = 'add-recipes'
    recipes_parser = subparsers.add_parser(recipes_cmd, help='add recipes from git remote')
    recipes_parser.add_argument('url', help='url of the remote (e.g. git@github.com:<username>/<reponame>.git or https://github.com/<username>/<reponame>.git)')
    recipes_parser.add_argument('--tag', '-t', required=False, default='master')
    recipes_parser.add_argument('--verbose', '-v', required=False, action='store_true', help='print additional information')
    recipes_parser.add_argument('--clone-protocol', required=False, choices=cloneprotos, help='override clone protocol')

    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    # initialize workspace
    if args.command == init_cmd:
        # create marker file
        write_ws_file(rootdir=rootdir)  # note: error on failure?

    if args.version:
        config = ConfigParser()
        src_path = Path(os.path.abspath(__file__)).parent.parent
        root_path = src_path.parent
        cfg_path = os.path.join(root_path, 'setup.cfg')
        config.read(cfg_path)
        try:
            version = f"hhcm-forest {config['metadata']['version']} ({src_path})"

        except KeyError:
            version = get_distribution('hhcm-forest')

        print(version)

        return True

    # check ws
    if not check_ws_file(rootdir=rootdir):
        print(f'current directory {rootdir} is not a forest workspace.. \
    have you called forest init ?', file=sys.stderr)
        return False

    # create directories (if do not exist)
    for dir in (buildroot, installdir, srcroot, recipesdir):
        if not os.path.exists(dir):
            os.mkdir(dir)

    # create setup.bash if does not exist
    write_setup_file()

    # verbose mode will show output of any called process
    if args.verbose:
        from forest.common import proc_utils
        proc_utils.call_process_verbose = True

    if not args.verbose:
        from forest.common import print_utils
        print_utils.log_file = open(args.log_file, 'w+')

    # sudo handling 
    if args.command == grow_cmd and args.pwd is not None:
        sudo_refresher = sudo_refresh.SudoRefresher(pwd=args.pwd)

    # print available packages
    if args.list:
        recipes = Cookbook.get_available_recipes()
        print(' '.join(recipes))
        return True

    # set config vars
    if args.command == grow_cmd and args.config:
        from forest.common import config_handler
        ch = config_handler.ConfigHandler.instance()
        ch.set_config_variables(args.config)

    # print available local attributes for conditional args
    if args.command == grow_cmd and args.list_eval_locals:
        from forest.common import eval_handler
        eval_handler.EvalHandler.print_available_locals()
        return True

    # clone proto
    if args.command == grow_cmd and args.clone_protocol is not None:
        from forest.common.fetch_handler import GitFetcher
        GitFetcher.proto_override = args.clone_protocol

    # clone proto
    if args.command == grow_cmd and args.clone_depth is not None:
        from forest.common.fetch_handler import GitFetcher
        GitFetcher.depth_override = args.clone_depth

    # if required, add a recipe repository to the list of remotes
    if args.command == recipes_cmd:
        print('adding recipes...')
        recipe_source = RecipeSource.FromUrl(args.url, args.tag, force_proto=args.clone_protocol)
        return Cookbook.add_recipes(recipe_source)

    # no recipe to install, exit
    if args.command == grow_cmd and args.recipe is None:
        print('no recipe to build, exiting...')
        return True

    # uninstall functionality
    if args.command == grow_cmd and args.uninstall:
        return uninstall_package(pkg=args.recipe,
                                 buildroot=buildroot,
                                 installdir=installdir,
                                 verbose=args.verbose)

    # clean functionality
    if args.command == grow_cmd and args.clean:
        return clean(pkg=args.recipe,
                     buildroot=buildroot,
                     installdir=installdir,
                     verbose=args.verbose)

    # handle modes
    if args.command == grow_cmd and args.mode is not None:
        EvalHandler.modes = set(args.mode)

    # default cmake args
    if args.command == grow_cmd and args.cmake_args:
        cmake_tools.CmakeTools.set_default_args(['-D' + a for a in args.cmake_args])

    # print jobs
    if args.command == grow_cmd:

        # check ws is sourced
        if rootdir not in os.environ.get('HHCM_FOREST_PATH', '').split(':'):
            print('[warn] forest workspace does not appear to be sourced')
        
        print(f'building {args.recipe} with {args.jobs} parallel job{"s" if int(args.jobs) > 1 else ""}')

        # perform required installation
        success = install_package(pkg=args.recipe,
                                  srcroot=srcroot,
                                  buildroot=buildroot,
                                  installdir=installdir,
                                  buildtype=args.default_build_type,
                                  jobs=args.jobs,
                                  reconfigure=args.force_reconfigure,
                                  no_deps=args.no_deps
                                  )

        return success

    return True


if __name__ == '__main__':
    main()


import os

rootdir = os.getcwd()
recipesdir = os.path.join(rootdir, 'recipes')
buildroot = os.path.join(rootdir, 'build')
installdir = os.path.join(rootdir, 'install')
srcroot = os.path.join(rootdir, 'src')


import multiprocessing
import psutil
import os
import subprocess
import time

from forest.common import proc_utils

class SudoRefresher:

    """
    SudoRefresher spawns a process that keeps sudo's session alive forever.
    """

    def __init__(self, pwd=None) -> None:

        # first call to sudo is done inside parent process (sync)
        SudoRefresher._do_sudo(pwd)
        print('')  # needed to add newline after sudo's prompt

        # then, start background process
        parent_pid = os.getpid()
        multiprocessing.set_start_method('spawn')
        self.proc = multiprocessing.Process(
                target=SudoRefresher._do_sudo_refresh, 
                args=(parent_pid, pwd)
                )
        self.proc.start()

    def _do_sudo(pwd):

        # password not provided, sudo's prompt will be used
        if pwd is None:
            args = ['sudo', '-v']  # sudo -v extends credentials 
            input = None
        else:
            args = ['sudo', '-S', '-v', '--prompt', '[sudo] got password']
            input = (pwd + '\n').encode()

        try:
            subprocess.check_output(args=args, input=input)
        except subprocess.CalledProcessError as e:
            print(f'could not call "sudo -v": {e.stderr}')


    def _do_sudo_refresh(parent_pid, pwd):

        if proc_utils.call_process_verbose:
            print(f'[sudo refreshing loop] started with pid {os.getpid()}')

        while psutil.pid_exists(parent_pid):
            SudoRefresher._do_sudo(pwd)
            time.sleep(60)
            
    
    def __del__(self):
        self.proc.terminate()  # we don't need clean termination
        self.proc.join(timeout=5)
        if self.proc.exitcode is None:
            print('warning: sudo refresher has not terminated yet')
        else:
            pass  # normal termination



import site
import sys

from setuptools import setup

if __name__ == "__main__":
    site.ENABLE_USER_SITE = "--user" in sys.argv[1:]
    setup(data_files=[
      ('share/bash-completion/completions/', ['src/forest/forest.bash']),
    ])


import os
import sys
import glob 
import subprocess

if __name__ == '__main__':

    # set https as default 
    os.putenv('HHCM_FOREST_CLONE_DEFAULT_PROTO', 'https')

    curr_dir = os.path.dirname(os.path.abspath(__file__))
    bash_test_files = glob.glob(curr_dir + '/test_*.bash')
    bash_test_files.sort()
    sep = '\n  - '
    print(f'found bash test files in {curr_dir}: {sep.join(bash_test_files)}')
    
    failed = False
    for f in bash_test_files:
        print(f'\n\n\n>>>>>>> running test {f}')
        try:
            subprocess.run(f, check=True)
        except subprocess.CalledProcessError as e:
            failed = True 
            print(e)

    if failed:
        sys.exit(1)
